/****************************************************************
 *  Vidalia is distributed under the following license:
 *
 *  Copyright (C) 2006,  Matt Edman, Justin Hipple
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 *  Boston, MA  02110-1301, USA.
 ****************************************************************/

/** 
 * \file vidaliasettings.cpp
 * \version $Id: vidaliasettings.cpp 1064 2006-06-27 00:40:05Z edmanm $
 */

#include <QDir>
#include <QCoreApplication>
#include <lang/languagesupport.h>
#include <control/logevent.h>
#include <vidalia.h>

#include "vidaliasettings.h"

#if defined(Q_WS_WIN)
#include <util/registry.h>
#include <util/win32.h>
#endif


/* Vidalia's Settings */
#define SETTING_LANGUAGE            "LanguageCode"
#define SETTING_STYLE               "InterfaceStyle"
#define SETTING_RUN_TOR_AT_START    "RunTorAtStart"
#define SETTING_DATA_DIRECTORY      "DataDirectory"

#define SETTING_MSG_FILTER          "MessageLog/MessageFilter"
#define SETTING_MAX_MESSAGE         "MessageLog/MaxMsgCount"
#define SETTING_ENABLE_LOG_FILE     "MessageLog/EnableLogFile"
#define SETTING_LOG_FILE            "MessageLog/LogFile"

#define SETTING_BWGRAPH_FILTER        "BandwidthGraph/BWLineFilter"
#define SETTING_BWGRAPH_OPACITY       "BandwidthGraph/Opacity"
#define SETTING_BWGRAPH_ALWAYS_ON_TOP "BandwidthGraph/AlwaysOnTop"

/* Default Vidalia Settings */
#if defined(Q_WS_MAC)
#define DEFAULT_STYLE               "macintosh (aqua)"
#else
#define DEFAULT_STYLE               "plastique"
#endif

#define DEFAULT_LANGUAGE            LanguageSupport::defaultLanguageCode()
#define DEFAULT_RUN_TOR_AT_START    false
#define DEFAULT_OPACITY             100

/* Default message log settings */
#define DEFAULT_MSG_FILTER          (LogEvent::Error|LogEvent::Warn|LogEvent::Notice)
#define DEFAULT_MAX_MESSAGE         250
#define DEFAULT_ENABLE_LOG_FILE     false

#if defined(Q_OS_WIN32)
#define DEFAULT_LOG_FILE       (win32_program_files_folder()+"\\Tor\\tor-log.txt")
#define STARTUP_REG_KEY        "Software\\Microsoft\\Windows\\CurrentVersion\\Run"
#define VIDALIA_REG_KEY        "Vidalia" 
#else
#define DEFAULT_LOG_FILE       (QDir::homePath() + "/.tor/tor-log.txt")
#endif

/* Default bandwidth graph settings */
#define DEFAULT_BWGRAPH_FILTER          (BWGRAPH_SEND|BWGRAPH_REC)
#define DEFAULT_BWGRAPH_ALWAYS_ON_TOP   false

/** The location of Vidalia's settings and configuration file. */
#define SETTINGS_FILE   (Vidalia::dataDirectory() + "/vidalia.conf")


/** Default Constructor */
VidaliaSettings::VidaliaSettings()
: QSettings(SETTINGS_FILE, QSettings::IniFormat)
{  
}

/** Resets all of Vidalia's settings. */
void
VidaliaSettings::reset()
{
  QSettings settings(SETTINGS_FILE, QSettings::IniFormat);
  settings.clear();
}

/** Gets the currently preferred language code for Vidalia. */
QString
VidaliaSettings::getLanguageCode()
{
  return value(SETTING_LANGUAGE, DEFAULT_LANGUAGE).toString();
}

/** Sets the preferred language code. */
void
VidaliaSettings::setLanguageCode(QString languageCode)
{
  setValue(SETTING_LANGUAGE, languageCode);
}

/** Gets the interface style key (e.g., "windows", "motif", etc.) */
QString
VidaliaSettings::getInterfaceStyle()
{
  return value(SETTING_STYLE, DEFAULT_STYLE).toString();
}

/** Sets the interface style key. */
void
VidaliaSettings::setInterfaceStyle(QString styleKey)
{
  setValue(SETTING_STYLE, styleKey);
}

/** Returns true if Tor is to be run when Vidalia starts. */
bool
VidaliaSettings::runTorAtStart()
{
  return value(SETTING_RUN_TOR_AT_START, DEFAULT_RUN_TOR_AT_START).toBool();
}

/** If <b>run</b> is set to true, then Tor will be run when Vidalia starts. */
void
VidaliaSettings::setRunTorAtStart(bool run)
{
  setValue(SETTING_RUN_TOR_AT_START, run);
}

/** Returns true if Vidalia is set to run on system boot. */
bool
VidaliaSettings::runVidaliaOnBoot()
{
#if defined(Q_WS_WIN)
  if (!registry_get_key_value(STARTUP_REG_KEY, VIDALIA_REG_KEY).isEmpty()) {
    return true;
  } else {
    return false;
  }
#else
  /* Platforms other than windows aren't supported yet */
  return false;
#endif
}

/** If <b>run</b> is set to true, then Vidalia will run on system boot. */
void
VidaliaSettings::setRunVidaliaOnBoot(bool run)
{
#if defined(Q_WS_WIN)
  if (run) {
    registry_set_key_value(STARTUP_REG_KEY, VIDALIA_REG_KEY,
        QString("\"" +
                QDir::convertSeparators(QCoreApplication::applicationFilePath())) +
                "\"");
  } else {
    registry_remove_key(STARTUP_REG_KEY, VIDALIA_REG_KEY);
  }
#else
  /* Platforms othe rthan windows aren't supported yet */
  Q_UNUSED(run);
  return;
#endif
}

/** Returns the current message filter. */
uint
VidaliaSettings::getMsgFilter()
{
  return value(SETTING_MSG_FILTER, DEFAULT_MSG_FILTER).toUInt(); 
}

/** Saves the setting for whether or not the given message severity will be
 * displayed. */
void
VidaliaSettings::setMsgFilter(LogEvent::Severity severity, bool add)
{
  uint filter = getMsgFilter();
  filter = (add ? (filter | (uint)severity) : (filter & ~((uint)severity)));
  setValue(SETTING_MSG_FILTER, filter);
}

/** Set maximum number of messages to display in log. */
void
VidaliaSettings::setMaxMsgCount(int max)
{
  setValue(SETTING_MAX_MESSAGE, max);
}

/** Return maximum number of messages to display in log. */
int
VidaliaSettings::getMaxMsgCount()
{
  return value(SETTING_MAX_MESSAGE, DEFAULT_MAX_MESSAGE).toInt();
}

/** Returns whether or not to enable automatically saving log messages from
 * the message log to a file on disk. */
bool
VidaliaSettings::isLogFileEnabled()
{
  return value(SETTING_ENABLE_LOG_FILE, 
               DEFAULT_ENABLE_LOG_FILE).toBool();
}

/** Sets whether or not to enable automatically saving log messages from the
 * message log to a file on disk. */
void
VidaliaSettings::enableLogFile(bool enable)
{
  setValue(SETTING_ENABLE_LOG_FILE, enable);
}

/** Gets the destination file on disk to which log messages can be saved. */
QString
VidaliaSettings::getLogFile()
{
  return QDir::convertSeparators(
                  value(SETTING_LOG_FILE, DEFAULT_LOG_FILE).toString());
}

/** Sets the destination file on disk to which log messages can be saved. */
void
VidaliaSettings::setLogFile(QString file)
{
  setValue(SETTING_LOG_FILE, QDir::convertSeparators(file));
}

/** Returns the bandwidth line filter. */
uint
VidaliaSettings::getBWGraphFilter()
{
  return value(SETTING_BWGRAPH_FILTER, DEFAULT_BWGRAPH_FILTER).toUInt(); 
}

/** Saves the setting for whether or not the given line will be graphed */
void
VidaliaSettings::setBWGraphFilter(uint line, bool status)
{
  uint filter = getBWGraphFilter();
  filter = (status ? (filter | line) : (filter & ~(line)));
  setValue(SETTING_BWGRAPH_FILTER, filter);
}

/** Get the level of opacity for the BandwidthGraph window. */
int
VidaliaSettings::getBWGraphOpacity()
{
  return value(SETTING_BWGRAPH_OPACITY, DEFAULT_OPACITY).toInt();
}

/** Set the level of opacity for the BandwidthGraph window. */
void
VidaliaSettings::setBWGraphOpacity(int value)
{
  setValue(SETTING_BWGRAPH_OPACITY, value);
}

/** Gets whether the bandwidth graph is always on top when displayed. */
bool
VidaliaSettings::getBWGraphAlwaysOnTop()
{
  return value(SETTING_BWGRAPH_ALWAYS_ON_TOP,
               DEFAULT_BWGRAPH_ALWAYS_ON_TOP).toBool();
}

/** Sets whether the bandwidth graph is always on top when displayed. */
void
VidaliaSettings::setBWGraphAlwaysOnTop(bool alwaysOnTop)
{
  setValue(SETTING_BWGRAPH_ALWAYS_ON_TOP, alwaysOnTop);
}

